`include "vhd_wishbone_master.svh"
`include "logger.svh"
`include "mock_turtle_driver.svh"

`include "masterfip_driver.svh"
`include "masterfip_macrocycle_config.svh"
`include "fip_phy_model.svh"
`include "fip_hardware_model.svh"
`include "fip_monitor_device.svh"
`include "fip_agent_device.svh"
`include "fip_virtual_master.svh"

class FipTestCaseBase extends LoggerClient;

   protected FipMonitorDevice m_mon;

   protected MasterfipDriver m_fip;
   protected MasterfipTestMacrocycle m_cycle;
   protected MasterfipTestAgentDevice m_agents[4];
   protected time m_startTime;
   protected int  m_speed;
   protected int  m_macrocycleVariant;
   protected MasterfipVariablePayload m_varPayloads[$];
   protected MasterfipVariablePayload m_msgPayloads[$];
   protected mstrfip_report m_report;
   protected bit  m_failed;
   protected virtual MasterfipMezzanine m_board;


   function automatic void setDriver ( MasterfipDriver drv );
      m_fip = drv;
   endfunction // setDriver
   
   function automatic void attachBoard( virtual MasterfipMezzanine brd );
      m_board = brd;
   endfunction // attachBoard

   function automatic void setSpeed( int speed );
      m_speed = speed;
   endfunction // setSpeed

   function automatic void setMacrocycleVariant ( int macrocycle );
      m_macrocycleVariant = macrocycle;
   endfunction // setMacrocycleVariant
   
   function new;
      m_failed = 0;
      m_speed = 2500000;
      m_macrocycleVariant = 0;
   endfunction 

   virtual function automatic string getName();
      return "?";
   endfunction // getName
   
   virtual task automatic configureAgents( virtual FipPHY mon, 
					   virtual FipPHY agent1,
					   virtual FipPHY agent2,
					   virtual FipPHY agent3
);
      m_mon = new;
      m_agents[1] = new( 'h01, m_macrocycleVariant );
      m_agents[2] = new( 'h5f, m_macrocycleVariant );
      m_agents[3] = new( 'hb4, m_macrocycleVariant );

      m_mon.attach( mon );
      m_agents[1].attach ( agent1 );
      m_agents[2].attach ( agent2 );
      m_agents[3].attach ( agent3 );

      m_mon.setSpeed( m_speed );
      m_agents[1].setSpeed( m_speed );
      m_agents[2].setSpeed( m_speed );
      m_agents[3].setSpeed( m_speed );

   
      
   endtask // configureAgents
   
   virtual task automatic setup();
      m_fip.reset();

      m_board.setSpeed(m_speed);
      
      m_cycle = new;
      m_cycle.setSpeed( m_speed );
      m_cycle.setVariant ( m_macrocycleVariant );
      m_cycle.configureMacrocycle( m_fip );
      m_fip.startMacrocycle();
      m_cycle.setPeriodicVariables( m_fip );
      m_cycle.setMessages( m_fip );
      #100us;
   endtask // setup

   virtual task automatic run();
      m_board.pulseTrigger();
      m_startTime = $time;
   endtask // run
   
   
   virtual function automatic bit isFinished();
      automatic time t;
      
      case(m_speed)
	2500000: t = 11ms;
	1000000: t = 17ms;
	31250: t = 500ms;
      endcase // case (m_speed)
      
      return ($time - m_startTime > t);
      
   endfunction // isFinished


   

   virtual task automatic validateRxFrame( MasterfipVariablePayload hostPayload, FipFrame refFrame, int agent );

      if( refFrame == null )
	begin
	   fail($sformatf("RefFrame == null? Did the master send it at all?"));
	   return;
	end
      
      if ( hostPayload.error )
	begin
	   fail($sformatf("Periodic Variable received by the MasterFIP with error (code 0x%x, agent 0x%x).", hostPayload.error, agent));
	   return;

      end
      
      if ( !refFrame.equals (hostPayload.asFipFrame(refFrame.getControl() ) )  )
	begin
	   fail($sformatf("Periodic Variable received by the MasterFIP doesn't correspond to the frame produced by the agent 0x%x.", agent));
	   return;
	end
   endtask // validateRxFrame

   virtual task automatic retrieveVarsFromHMQ( int var_count, int msg_count );
      automatic int status;

      m_varPayloads = m_fip.getVariablePayloads();
      m_msgPayloads = m_fip.getMessagePayloads();

      $display("m %d p %d\n", m_msgPayloads.size(), m_varPayloads.size() );
      
      
      if ( var_count && m_varPayloads.size() != var_count )
	begin
      	   fail($sformatf("Expected %-0d periodic variable payloads in HMQ, got %-0d", var_count, m_varPayloads.size()));
	   m_failed = 1;
	end

      if ( msg_count && m_msgPayloads.size() != msg_count )
	begin
      	   fail($sformatf("Expected %-0d periodic message payloads in HMQ, got %-0d", msg_count, m_msgPayloads.size()));
	   m_failed = 1;
	end
      
   endtask // retrieveConsumedVariables
   
   virtual task automatic readReport( );
      automatic int status;

      m_report = new;
      m_fip.getReport( m_report );
      
      msg(0, $sformatf("Checking MasterFIP Status Report:") );
      msg(0, $sformatf("TX OK:                %-0d", m_report.tx_ok) );
      msg(0, $sformatf("TX Errors:            %-0d", m_report.tx_err) );
      msg(0, $sformatf("RX OK:                %-0d", m_report.rx_ok) );
      msg(0, $sformatf("RX Errors:            %-0d", m_report.rx_err) );
      msg(0, $sformatf("RX Timeouts:          %-0d", m_report.rx_tmo) );
      msg(0, $sformatf("Executed cycles:      %-0d", m_report.cycles) );
      msg(0, $sformatf("Sync Pulses:          %-0d", m_report.ext_sync_pulse_count) );
      msg(0, $sformatf("Fieldrive TX Errors:  %-0d", m_report.fd_tx_err ) );
      msg(0, $sformatf(" - cycle:             %-0d", m_report.fd_tx_err_cycle ) );
      msg(0, $sformatf(" - hwtime:            %-0d", m_report.fd_tx_err_hwtime ) );
      msg(0, $sformatf("Fieldrive WD Errors:  %-0d", m_report.fd_tx_watchdog ) );
      msg(0, $sformatf(" - cycle:             %-0d", m_report.fd_tx_watchdog_cycle ) );
      msg(0, $sformatf(" - hwtime:            %-0d", m_report.fd_tx_watchdog_hwtime ) );
      msg(0, $sformatf("RX MPS Status Errors: %-0d", m_report.rx_mps_status_err ) );
      msg(0, $sformatf("TX MPS Status Errors: %-0d", m_report.tx_mps_status_err ) );
      
   endtask

  
   virtual task automatic expectTxOk( int x );
      if ( m_report.tx_ok != x )
	begin
	   fail($sformatf("Expected TX OK = %-0d in the report, got %-0d", x, m_report.tx_ok ) );
	   m_failed = 1;
	end
   endtask // expectTxOk

   virtual task automatic expectRxOk( int x );
      if ( m_report.rx_ok != x )
	begin
	   fail($sformatf("Expected RX OK = %-0d in the report, got %-0d", x, m_report.rx_ok ) );
	   m_failed = 1;
	end
   endtask

   virtual task automatic expectRxErrors( int x );
      if ( m_report.rx_err != x )
	begin
	   fail($sformatf("Expected RX Errors = %-0d in the report, got %-0d", x, m_report.rx_err ) );
	   m_failed = 1;
	end
   endtask // expectRxErrors

   virtual task automatic expectRxTimeouts( int x );
      if ( m_report.rx_tmo != x )
	begin
	   fail($sformatf("Expected RX Timeouts = %-0d in the report, got %-0d", x, m_report.rx_tmo ) );
	   m_failed = 1;
	end
   endtask // expectRxErrors
   
   virtual task automatic expectTxErrors( int x );
      if ( m_report.tx_err != x )
	begin
	   fail($sformatf("Expected TX Errors = %-0d in the report, got %-0d", x, m_report.tx_err ) );
	   m_failed = 1;
	end
   endtask
   
   virtual task automatic expectMPSTxErrors( int x );
      if ( m_report.tx_mps_status_err != x )
	begin
	   fail($sformatf("Expected TX MPS Status Errors = %-0d in the report, got %-0d", x, m_report.tx_mps_status_err ) );
	   m_failed = 1;
	end
   endtask // expectMPSTxErrors

   virtual task automatic expectFieldriveTxErrors( int minval, int maxval );
      if ( m_report.fd_tx_err < minval || m_report.fd_tx_err > maxval )
	begin
	   fail($sformatf("Expected Fieldrive TX Errors >= %-0d and <= %-0d in the report, got %-0d", minval, maxval, m_report.fd_tx_err ) );
	   m_failed = 1;
	end
   endtask // expectMPSTxErrors

   virtual task automatic expectFieldriveWatchdogErrors( int minval, int maxval );
      if ( m_report.fd_tx_watchdog < minval || m_report.fd_tx_watchdog > maxval )
	begin
	   fail($sformatf("Expected Fieldrive Watchdog Errors >= %-0d and <= %-0d in the report, got %-0d", minval, maxval, m_report.fd_tx_watchdog ) );
	   m_failed = 1;
	end
   endtask // expectMPSTxErrors

   virtual task automatic expectMPSRxErrors( int x );
      if ( m_report.rx_mps_status_err != x )
	begin
	   fail($sformatf("Expected RX MPS Status Errors = %-0d in the report, got %-0d", x, m_report.rx_mps_status_err ) );
	   m_failed = 1;
	end
   endtask

   
   virtual task automatic expectMessagePayload ( int hmqIndex, int agent, int expectError = 0 );
      automatic FipFrame refPayload = m_agents[agent].getProducedMsg('h02);
      automatic MasterfipVariablePayload hostPayload = m_msgPayloads[ hmqIndex ];
      automatic FipFrame t = hostPayload.asFipFrame(refPayload.getControl() );
      
      
      msg(2, $sformatf("Checking message produced by agent %-0d (expected error code 0x%-0x)", agent, expectError ) );
      
      if( hostPayload == null )
	begin
	   fail($sformatf("HostPayload == null? Did the master send it to HMQ at all?"));
	   return;
	end

      if( expectError )
	begin
	   if( hostPayload.error != expectError )
	     begin
		fail($sformatf("Expected error code 0x%x in the HMQ for message produced by agent %-0d, got 0x%x instead", expectError, agent, hostPayload.error ) );
		return;
	     end

	   return;
	end
      
      if ( hostPayload.error && !expectError )
	begin
	   fail($sformatf("Aperiodic message received by the MasterFIP with error (code 0x%x, agent 0x%x).", hostPayload.error, agent));
	   return;
      end
      
      if ( !refPayload.equals (hostPayload.asFipFrame(refPayload.getControl() ) )  )
	begin
	   fail($sformatf("Aperiodic message received by the MasterFIP doesn't correspond to the frame produced by the agent %-0d.", agent) );
	   return;
	end

      
   endtask // expectMessagePayload
   
   virtual task automatic expectVariablePayload ( int hmqIndex, int agent, int expectError = 0 );

      automatic FipFrame refPayload = m_agents[agent].getProducedVar('h02);
      automatic MasterfipVariablePayload hostPayload = m_varPayloads[ hmqIndex ];

      msg(2, $sformatf("Checking variable produced by agent %-0d (expected error code 0x%-0x)", agent, expectError ) );
      
      if( hostPayload == null )
	begin
	   fail($sformatf("HostPayload == null? Did the master send it to HMQ at all?"));
	   return;
	end

      if( expectError )
	begin
	   if( hostPayload.error != expectError )
	     fail($sformatf("Expected error code 0x%x in the HMQ for variable produced by agent %-0d, got 0x%x instead", expectError, agent, hostPayload.error ) );
	   msg(2, "Error codes match");

	   return;
	end
      
      if ( hostPayload.error && !expectError )
	begin
	   fail($sformatf("Periodic Variable received by the MasterFIP with error (code 0x%x, agent 0x%x).", hostPayload.error, agent));
	   return;
      end
      
      if ( !refPayload.equals (hostPayload.asFipFrame(refPayload.getControl() ) )  )
	begin
	   fail($sformatf("Periodic Variable received by the MasterFIP doesn't correspond to the frame produced by the agent %-0d.", agent) );
	   return;
	end

      
   endtask // expectVariablePayloar

   virtual function automatic bit expectConsumedVariable ( int agent, int dataRange );
      automatic FipFrame consumedFrame = m_agents[agent].getConsumedVar('h01);

      msg(2, $sformatf("Checking variable consumed by agent %-0d", agent ) );
      
      if( !consumedFrame.isCRCCorrect() )
	begin
	   fail($sformatf("Variable consumed by agent %-0d has incorrect CRC.", agent));
	   return 0;
	end

      if ( !consumedFrame.equals( FipFrame::makeRP_DAT( FT_RP_DAT, makeRange(0, dataRange) ) ) )
	begin
	   fail($sformatf("TX Variable [agent %-0d]: incorrect payload.", agent));
	   return 0;
	end
      return 1;
   endfunction // expectConsumedVariable
   
   virtual task automatic validateVariables();
   endtask // validateVariables

   virtual task automatic validateSMMPS();
      automatic mstrfip_presence_report p_list = new;
      automatic int status;
      automatic FipFrameList frames = m_mon.getFrames();
      automatic int lastAgent = -1, presenceCount = 0, presenceAnswerCount = 0;
      automatic int expectedPresenceCount = 0;
      
      uint32_t chk[$]= '{'h1, 0, 0, 0, 0, 0, 0};
      

      msg(2, "Verifying node presence map");
      
      m_fip.getPresentList( p_list, status );
      
      for(int i = 0 ;i < 8;i++)
	if ( p_list.presenceMap[i] != chk[i])
	  begin
	     fail( $sformatf("Expected presenceMap[%-0d] = 0x%x, got 0x%x", i, chk[i], p_list.presenceMap[i] ) );
	     m_failed = 1;
	     return;
	  end

      msg(2, "Verifying presence frames");
      
      foreach(frames[i])
	begin
	   automatic FipFrame f = frames[i];

	   if( f.getControl() == FT_ID_DAT && f.getVariable() == 'h14 )
	     begin
		presenceCount++;
		lastAgent = f.getAddress();
	     end
	   else if ( f.getControl() == FT_RP_DAT_RQ && lastAgent == 1 ) begin
	      presenceAnswerCount++;
	   end else begin
	      lastAgent = -1;
	   end
	end
      
      case (m_speed)
	2500000: expectedPresenceCount = 20;
	1000000: expectedPresenceCount = 17;
	31250: expectedPresenceCount = 0;
      endcase // case (m_speed)
      
      if ( presenceAnswerCount != 1 )
	begin
	   fail($sformatf("Expected 1 presence request answer frames, got %-0d", presenceAnswerCount ) );
	   return;
	end

      if ( presenceCount != expectedPresenceCount )
	begin
	   fail($sformatf("Expected %-0d presence probe frames, got %-0d", expectedPresenceCount, presenceCount ) );
	   return;
	end
      
	

   endtask // validateSMMPS

   virtual task automatic validateReport();
   endtask // validateREport

   virtual task automatic validate();
      validateVariables();

      if (m_failed)
	return;

      validateReport();
      validateSMMPS();
   endtask // validate
   
endclass // FipTestCaseBase


class FipTCFullMacrocycle extends FipTestCaseBase;

   virtual task automatic validateVariables();
      retrieveVarsFromHMQ(3, m_macrocycleVariant == 0 ? 1 : 2);
            
      expectConsumedVariable( 1, 'h3f );
      expectConsumedVariable( 2, 'h1 );
      expectConsumedVariable( 3, 'h7b );

      expectVariablePayload( 0, 1 ); // first comes agent 1
      expectVariablePayload( 1, 2 ); // then comes agent 2
      expectVariablePayload( 2, 3 ); // ...then agent 3

      expectMessagePayload( 0, 1 );
      if (m_macrocycleVariant == 1)
	expectMessagePayload( 1, 2 );
      
   endtask // validateVariables
   
   virtual task automatic validateReport();
      readReport();
      expectTxErrors(0);
      expectRxOk(m_macrocycleVariant == 0 ? 6 : 8);
      expectRxErrors(0);
      expectMPSRxErrors(0);
      expectMPSTxErrors(0);
   endtask // validateREport

   function automatic string getName();
      return $sformatf ("Run Full Macrocycle (variant %-0d), speed %-0d bps.", m_macrocycleVariant, m_speed );
   endfunction // getName
   
endclass // FipTCFullMacrocycle



// Reception of RP_DAT with an incorrect CTRL byte
class FipTC_RP_DAT_Wrong_Ctrl extends FipTestCaseBase;

   protected uint8_t m_controlValue;
   
   function new ( uint8_t controlValue );
      super.new( );
      m_controlValue = controlValue;
   endfunction // new
   
   function automatic string getName();
      return $sformatf("Reception of RP_DAT with incorrect CTRL byte [0x%x]", m_controlValue );
   endfunction // getName

   task automatic setup();
      automatic FipFrame frame;

      super.setup();
      frame = m_agents[2].getProducedVar('h02);
      frame.forceControlField( 1, m_controlValue ); // dummy CTRL byte
   endtask // setup

   virtual task automatic validateVariables();
      retrieveVarsFromHMQ( 3, 1 );
      expectConsumedVariable( 1, 'h3f );
      expectConsumedVariable( 2, 'h1 );
      expectConsumedVariable( 3, 'h7b );
      expectVariablePayload( 0, 1 ); // first comes agent 1
      expectVariablePayload( 1, 2, `MSTRFIP_FRAME_BAD_CTRL  ); // then comes agent 2
      expectVariablePayload( 2, 3 ); // ...then agent 3
      expectMessagePayload( 0, 1 );
   endtask // validateVariables

   virtual task automatic validateReport();
      readReport();
      expectTxErrors(0);
      expectRxOk(6);
      expectRxErrors(0);
      expectMPSRxErrors(0);
      expectMPSTxErrors(0);
   endtask // validateREport
   
endclass



// Reception of RP_DAT with an incorrect PDU byte
class FipTC_RP_DAT_Wrong_PDU extends FipTestCaseBase;

   protected uint8_t m_pduValue;
   
   function new ( uint8_t pduValue );
      super.new();
      m_pduValue = pduValue;
   endfunction // new
   
   function automatic string getName();
      return $sformatf("Reception of RP_DAT with incorrect PDU byte [0x%x]", m_pduValue );
   endfunction // getName

   task automatic setup();
      automatic FipFrame frame;

      super.setup();
      frame = m_agents[2].getProducedVar('h02);
      frame.forcePDUField( 1, m_pduValue ); // override PDU value
   endtask // setup

   virtual task automatic validateVariables();
      retrieveVarsFromHMQ( 3, 1 );
      expectConsumedVariable( 1, 'h3f );
      expectConsumedVariable( 2, 'h1 );
      expectConsumedVariable( 3, 'h7b );
      expectVariablePayload( 0, 1 ); // first comes agent 1
      expectVariablePayload( 1, 2, `MSTRFIP_FRAME_BAD_PDU  ); // then comes agent 2
      expectVariablePayload( 2, 3 ); // ...then agent 3
      expectMessagePayload( 0, 1 );
   endtask // validateVariables

   virtual task automatic validateReport();
      readReport();
      expectTxErrors(0);
      expectRxOk(6);
      expectRxErrors(1);
      expectMPSRxErrors(0);
      expectMPSTxErrors(0);
   endtask // validateREport
   
endclass // FipTC_RP_DAT_Wrong_PDU

// Reception of RP_DAT with an incorrect MPS byte
class FipTC_RP_DAT_WrongMPS extends FipTestCaseBase;

   int m_variant;
   
   
   function new ( int variant );
      super.new();
      m_variant = variant;
   endfunction // new
   
   function automatic string getName();
      case (m_variant)
	0:
	  return "Reception of RP_DAT with wrong MPS byte"  ;
	1:
	  return "Reception of RP_DAT with non-fresh MPS status"  ;
	endcase
   endfunction // getName

   task automatic setup();
      automatic FipFrame frame;

      super.setup();
      frame = m_agents[2].getProducedVar('h02);
      case(m_variant)
	0: 
	  frame.setMPSStatus('h00 );
	1: 
	  frame.setMPSStatus('h04 );
      endcase // case (m_variant)
      
   endtask // setup

   virtual task automatic validateVariables();
      retrieveVarsFromHMQ( 3, 1 );
      expectConsumedVariable( 1, 'h3f );
      expectConsumedVariable( 2, 'h1 );
      expectConsumedVariable( 3, 'h7b );
      expectVariablePayload( 0, 1 ); // first comes agent 1
      expectVariablePayload( 1, 2 ); // then comes agent 2
      expectVariablePayload( 2, 3 ); // ...then agent 3
      expectMessagePayload( 0, 1 );
   endtask // validateVariables

   virtual task automatic validateReport();
      readReport();
      expectTxErrors(0);
      expectRxOk(6);
      expectRxErrors(0);
      expectMPSRxErrors(1);
      expectMPSTxErrors(0);
   endtask // validateREport
   
endclass


// Reception of RP_DAT with an incorrect PDU byte
class FipTC_RP_DAT_Wrong_Preamble extends FipTestCaseBase;

   protected int m_variant;
   
   function new ( int variant );
      super.new();
      m_variant = variant;
   endfunction // new
   
   function automatic string getName();
      case(m_variant)
	0: 
	  return "Reception of RP_DAT with a wrong preamble";
	1: 
	  return "Reception of Preamble with less than 8 bits but more than 4";
	2: 
	  return "Reception of Preamble with less than 4 bits";
      endcase // case (m_variant)
   endfunction // getName

   task automatic setup();
      automatic FipFrame frame;

      super.setup();

      frame = m_agents[2].getProducedVar('h02);

      case(m_variant)
	0:
	  frame.forcePreambleError(1);
	1:
	  frame.setPreambleBitCount(5);
	2:
	  frame.setPreambleBitCount(2);
      endcase // case (m_variant)
   endtask // setup

   virtual task automatic validateVariables();
      retrieveVarsFromHMQ( 3, 1 );
      expectConsumedVariable( 1, 'h3f );
      expectConsumedVariable( 2, 'h1 );
      expectConsumedVariable( 3, 'h7b );
      expectVariablePayload( 0, 1 ); // first comes agent 1
      // only 4-8 bits preamble should report no error

      expectVariablePayload( 1, 2, m_variant == 0 || m_variant == 2 ? `MSTRFIP_FRAME_ERR : 0 ); // then comes agent 2. truncated preamble (5 bits is assumed OK)
      expectVariablePayload( 2, 3 ); // ...then agent 3
      expectMessagePayload( 0, 1 );
      
   endtask // validateVariables

   virtual task automatic validateReport();
      readReport();
      expectTxErrors(0);
      // only 4-8 bits preamble should report no error
      expectRxOk(m_variant == 1 ? 6 : 5);
      expectRxErrors(m_variant == 1 ? 0 : 1);
      expectMPSRxErrors(0);
      expectMPSTxErrors(0);
   endtask // validateREport
endclass


// Reception of RP_DAT in less than turnaround time
class FipTC_RP_DAT_TurnaroundViolation extends FipTestCaseBase;

   
   function automatic string getName();
      return "Reception of Preamble in less time than the turnaround time";
   endfunction // getName

   task automatic setup();
      automatic FipFrame frame;
      super.setup();
      frame = m_agents[2].getProducedVar('h02);
      frame.forceTxTimeOffset(1, -m_agents[2].getTurnaroundTime() + 3us );
   endtask // setup

   virtual task automatic validateVariables();
      retrieveVarsFromHMQ( 3, 1 );
      expectConsumedVariable( 1, 'h3f );
      expectConsumedVariable( 2, 'h1 );
      expectConsumedVariable( 3, 'h7b );
      // MasteFIP is fast enough to accept ~2us of turnaround->no error should be reported
      expectVariablePayload( 0, 1 ); // first comes agent 1
      expectVariablePayload( 1, 2 );
      expectVariablePayload( 2, 3 ); // ...then agent 3
      expectMessagePayload( 0, 1 );
   endtask // validateVariables


    virtual task automatic validateReport();
      readReport();
      expectTxErrors(0);
      expectRxOk(6);
      expectRxErrors(0);
      expectMPSRxErrors(0);
      expectMPSTxErrors(0);
   endtask // validateREport
   
endclass


// Reception of RP_DAT in more than silence time
class FipTC_RP_DAT_SentAfterSilence extends FipTestCaseBase;

   
   
   function automatic string getName();
      return $sformatf("Reception of Preamble in more time than the silence time" );
   endfunction // getName

   task automatic setup();
      automatic FipFrame frame;

      super.setup();

      frame = m_agents[2].getProducedVar('h02);
      frame.forceTxTimeOffset(1, m_agents[2].getSilenceTime() );
   endtask // setup

  virtual task automatic validateVariables();
      retrieveVarsFromHMQ( 3, 1 );
      expectConsumedVariable( 1, 'h3f );
      expectConsumedVariable( 2, 'h1 );
      expectConsumedVariable( 3, 'h7b );
      expectVariablePayload( 0, 1 ); // first comes agent 1
     // ID_DAT sent to agent 3 is overwritten by the late response from agent 2 -> agent 2
     // gets an error (due to rubbish on the FIP bus), agent 3 a timeout (never receives an ID_DAT)
      expectVariablePayload( 1, 2, `MSTRFIP_FRAME_ERR );
      expectVariablePayload( 2, 3, `MSTRFIP_FRAME_TMO ); // ...then agent 3
      expectMessagePayload( 0, 1 );
   endtask // validateVariables


    virtual task automatic validateReport();
      readReport();
      expectTxErrors(0);
      expectRxOk(4);
      expectRxErrors(1);
      expectRxTimeouts(1);
      expectFieldriveTxErrors(10, 1000);
      expectMPSRxErrors(0);
      expectMPSTxErrors(0);
   endtask // validateREport
endclass


class FipTC_RP_DAT_WrongFES extends FipTestCaseBase;

   protected int m_variant;
   
   function new ( int variant );
      super.new( );
      m_variant = variant;
   endfunction // new
   
   function automatic string getName();
      case (m_variant)
	0: 
	  return "Reception of incorrect FES";
	1:
	  return "Reception of FES with less bits (7 bits)";
      endcase
   endfunction // getName

   task automatic setup();
      automatic FipFrame frame;

      super.setup();
      frame = m_agents[2].getProducedVar('h02);

      case(m_variant)
	0:
	  frame.forceFESError(1);
	1:
	  frame.forceTruncatedFES(1, 7);
      endcase // case (m_variant)
   endtask // setup

   virtual task automatic validateVariables();
      retrieveVarsFromHMQ( 3, 1 );
      expectConsumedVariable( 1, 'h3f );
      expectConsumedVariable( 2, 'h1 );
      expectConsumedVariable( 3, 'h7b );
      expectVariablePayload( 0, 1 ); // first comes agent 1
      expectVariablePayload( 1, 2, `MSTRFIP_FRAME_BAD_BSZ );
      expectVariablePayload( 2, 3 ); // ...then agent 3
      expectMessagePayload( 0, 1 );
   endtask // validateVariables


   virtual task automatic validateReport();
      readReport();
      expectTxErrors(0);
      expectRxOk(5);
      expectRxErrors(1);
      expectMPSRxErrors(0);
      expectMPSTxErrors(0);
   endtask // validateREport

endclass




class FipTC_RP_DAT_WrongCRC extends FipTestCaseBase;
   
   function automatic string getName();
      return "Reception of a RP_DAT with incorrect CRC";
   endfunction // getName

   task automatic setup();
      automatic FipFrame frame;

      super.setup();
      frame = m_agents[2].getProducedVar('h02);
      frame.forceIncorrectCRC(1);
   endtask // setup
   
   virtual task automatic validateVariables();
      retrieveVarsFromHMQ( 3, 1 );
      expectConsumedVariable( 1, 'h3f );
      expectConsumedVariable( 2, 'h1 );
      expectConsumedVariable( 3, 'h7b );
      expectVariablePayload( 0, 1 ); // first comes agent 1
      expectVariablePayload( 1, 2, `MSTRFIP_FRAME_ERR );
      expectVariablePayload( 2, 3 ); // ...then agent 3
      expectMessagePayload( 0, 1 );
   endtask // validateVariables


   virtual task automatic validateReport();
      readReport();
      expectTxErrors(0);
      expectRxOk(5);
      expectRxErrors(1);
      expectMPSRxErrors(0);
      expectMPSTxErrors(0);
   endtask // validateREport

endclass

class FipTC_RP_DAT_IncorrectLength extends FipTestCaseBase;

   protected int m_variant;
   
   function new ( int variant );
      m_variant = variant;
   endfunction // new
   
   function automatic string getName();
      case (m_variant)
	0:
	  return "Reception of RP_DAT too long (more than 124 bytes)";
	
	1: 
	  return "Reception of RP_DAT with length in bits not multiple of 8 but correct CRC";
	
	2: 
	  return "Reception of RP_DAT with length in bits not multiple of 8 and wrong CRC";

	3:
	  return "Reception of RP_DAT too short frame (zero data bytes)";
	
	4:
	  return "Reception of RP_DAT with LGHT byte not corresponding to the actual length";
      endcase // case (m_variant)
   endfunction // getName

   task automatic setup();
      automatic FipFrame frame;

      super.setup();
      frame = m_agents[2].getProducedVar('h02);

      case(m_variant)
	0:
	  m_agents[2].setProducedVar( 'h02, FipFrame::makeRP_DAT( FT_RP_DAT, makeRange(0, 'h80) ) );
	1:
	  frame.forceDataTxBitCount( 1, 6 * 8 - 1 );
	2:
	  begin
	     frame.forceDataTxBitCount( 1, 6 * 8 - 1 );
	     frame.forceIncorrectCRC(1);
	  end
	3:
	  m_agents[2].setProducedVar( 'h02, FipFrame::makeRP_DAT( FT_RP_DAT, '{} ) );
	4:
	  frame.forceLgthFieldValue( 1, 120 );
	endcase
      
   endtask // setup

   virtual task automatic validateVariables();

      case(m_variant)
	0:
	  begin
	     retrieveVarsFromHMQ(3, 0);
	     expectConsumedVariable( 1, 'h3f );
	     expectConsumedVariable( 2, 'h1 );
	     expectConsumedVariable( 3, 'h7b );
	     expectVariablePayload( 0, 1 ); // first comes agent 1
	     // 2nd and 3rd RP_DATs are never received due to bus conflict
	     expectVariablePayload( 1, 2, `MSTRFIP_FRAME_BAD_BSZ );
	     expectVariablePayload( 2, 3, `MSTRFIP_FRAME_ERR ); // ...then agent 3
	  end
	1:
	  begin
	     retrieveVarsFromHMQ(3, 1);
	     expectConsumedVariable( 1, 'h3f );
	     expectConsumedVariable( 2, 'h1 );
	     expectConsumedVariable( 3, 'h7b );
	     expectVariablePayload( 0, 1 ); // first comes agent 1
	     expectVariablePayload( 1, 2, `MSTRFIP_FRAME_ERR );
	     expectVariablePayload( 2, 3 ); // ...then agent 3
	     expectMessagePayload( 0, 1 );
	  end
	2:
	  begin
	     retrieveVarsFromHMQ(3, 1);
	     expectConsumedVariable( 1, 'h3f );
	     expectConsumedVariable( 2, 'h1 );
	     expectConsumedVariable( 3, 'h7b );
	     expectVariablePayload( 0, 1 ); // first comes agent 1
	     expectVariablePayload( 1, 2, `MSTRFIP_FRAME_ERR );
	     expectVariablePayload( 2, 3 ); // ...then agent 3
	     expectMessagePayload( 0, 1 );
	  end
	3:
	  begin
	     retrieveVarsFromHMQ(3, 1);
	     expectConsumedVariable( 1, 'h3f );
	     expectConsumedVariable( 2, 'h1 );
	     expectConsumedVariable( 3, 'h7b );
	     expectVariablePayload( 0, 1 ); // first comes agent 1
	     expectVariablePayload( 1, 2, `MSTRFIP_FRAME_BAD_BSZ );
	     expectVariablePayload( 2, 3 ); // ...then agent 3
	     expectMessagePayload( 0, 1 );
	  end	
	4:
	  begin
	     retrieveVarsFromHMQ(3, 1);
	     expectConsumedVariable( 1, 'h3f );
	     expectConsumedVariable( 2, 'h1 );
	     expectConsumedVariable( 3, 'h7b );
	     expectVariablePayload( 0, 1 ); // first comes agent 1
	     expectVariablePayload( 1, 2, `MSTRFIP_FRAME_BAD_BSZ );
	     expectVariablePayload( 2, 3 ); // ...then agent 3
	     expectMessagePayload( 0, 1 );
	  end
	
	
	endcase // case (m_variant)
      
	

   endtask // validateVariables

   virtual task automatic validateReport();

      readReport();
      case(m_variant)
	0: 
	  begin
	     expectTxErrors(0);
	     expectRxOk(2);
	     expectRxErrors(5);
	     expectMPSRxErrors(0);
	     expectMPSTxErrors(0);
	     expectFieldriveTxErrors(50, 1000);

	     // fixme: all report cases
	  end
	1: 
	  begin
	     expectTxErrors(0);
	     expectRxOk(5);
	     expectRxErrors(1);
	     expectMPSRxErrors(0);
	     expectMPSTxErrors(0);

	     // fixme: all report cases
	  end
	2: 
	  begin
	     expectTxErrors(0);
	     expectRxOk(5);
	     expectRxErrors(1);
	     expectMPSRxErrors(0);
	     expectMPSTxErrors(0);

	     // fixme: all report cases
	  end
	3: 
	  begin
	     expectTxErrors(0);
	     expectRxOk(6);
	     expectRxErrors(1);
	     expectMPSRxErrors(0);
	     expectMPSTxErrors(0);

	     // fixme: all report cases
	  end
	4: 
	  begin
	     expectTxErrors(0);
	     expectRxOk(6);
	     expectRxErrors(1);
	     expectMPSRxErrors(0);
	     expectMPSTxErrors(0);

	     // fixme: all report cases
	  end
      endcase // case (m_variant)
      
   endtask // validateREport
   
   
endclass // FipTC_RP_DAT_IncorrectLength



class FipTC_RP_DAT_FieldriveError extends FipTestCaseBase;

   protected int m_variant;
   
   function new ( int variant );
      m_variant = variant;
   endfunction // new
   
   function automatic string getName();
      case (m_variant)
	0:
	  return "Fieldrive TXER signal activated on transmission";
	
	1: 
	  return "Fieldrive WDGN signal activated on transmission";
      endcase // case (m_variant)
   endfunction // getName

   task automatic setup();
      super.setup();
   endtask // setup

   virtual task automatic run();
      super.run();
      #164us; // it comes in the middle of the 1st RP_DAT sent by the master @ 2.5 Mbps
      
      case(m_variant)
	0:
	  begin
	     m_board.U_Fieldrive.forceTxError(1);
	     #100us; 
	     m_board.U_Fieldrive.forceTxError(0);
	  end
	1:
	  begin
	     m_board.U_Fieldrive.forceWatchdogError();
	  end
      endcase // case (m_variant)
      
   endtask // run


   virtual task automatic validateVariables();
      case(m_variant)
	0:
	  begin
	     retrieveVarsFromHMQ(3, 0);
	     expectConsumedVariable( 2, 'h1 );
	     expectConsumedVariable( 3, 'h7b );
	     expectVariablePayload( 0, 1 ); // first comes agent 1
	     // 2nd and 3rd RP_DATs are never received due to bus conflict
	     expectVariablePayload( 1, 2 );
	     expectVariablePayload( 2, 3 ); // ...then agent 3
	  end
	1:
	  begin
	     retrieveVarsFromHMQ(3, 1);
	     expectConsumedVariable( 2, 'h1 );
	     expectConsumedVariable( 3, 'h7b );
	     expectVariablePayload( 0, 1 ); // first comes agent 1
	     expectVariablePayload( 1, 2 );
	     expectVariablePayload( 2, 3 ); // ...then agent 3
	     expectMessagePayload( 0, 1 );
	  end
	endcase // case (m_variant)

   endtask // validateVariables

   virtual task automatic validateReport();

      readReport();
      case(m_variant)
	0: 
	  begin
	     expectTxErrors(0);
	     expectRxOk(6);
	     expectRxErrors(0);
	     expectMPSRxErrors(0);
	     expectMPSTxErrors(0);
	     expectFieldriveTxErrors(1, 1);
	     expectFieldriveWatchdogErrors(0, 0);
	  end

	1:
	  begin
	     expectTxErrors(0);
	     expectRxOk(6);
	     expectRxErrors(0);
	     expectMPSRxErrors(0);
	     expectMPSTxErrors(0);
	     expectFieldriveTxErrors(0, 0);
	     expectFieldriveWatchdogErrors(1, 1);
	  end

      endcase // case (m_variant)
      
   endtask // validateREport
   
   
endclass // FipTC_RP_DAT_FieldriveError




// Reception of RP_DAT stopped in the middle of the frame
class FipTC_RP_DAT_PrematureEnd extends FipTestCaseBase;

   function new;
      super.new();
   endfunction // new
   
   function automatic string getName();
      return "Reception of incomplete RP_DAT frame (stopped in the middle of a frame)";
   endfunction // getName

   task automatic setup();
      automatic FipFrame frame;

      super.setup();
      frame = m_agents[3].getProducedVar('h02);
      frame.forceDataTxBitCount( 1, 64 * 8 ); // terminate the frame in the middle
      frame.forcePrematureEnd(1);
   endtask // setup

   virtual task automatic validateVariables();
      retrieveVarsFromHMQ( 3, 1 );
      expectConsumedVariable( 1, 'h3f );
      expectConsumedVariable( 2, 'h1 );
      expectConsumedVariable( 3, 'h7b );
      expectVariablePayload( 0, 1 ); // first comes agent 1
      expectVariablePayload( 1, 2 ); // then comes agent 2
      expectVariablePayload( 2, 3, `MSTRFIP_FRAME_BAD_BSZ ); // ...then agent 3
      expectMessagePayload( 0, 1 );
   endtask // validateVariables

   virtual task automatic validateReport();
      readReport();
      expectTxErrors(0);
      expectRxOk(5);
      expectRxErrors(1);
      expectMPSRxErrors(0);
      expectMPSTxErrors(0);
   endtask // validateREport
   
endclass // FipTC_RP_DAT_PrematureEnd

class FipTC_APP_VariableNotUpdated extends FipTestCaseBase;

   function new;
      super.new();
   endfunction // new
   
   function automatic string getName();
      return "Application not accessing the data to produce for a variable";
   endfunction // getName


   virtual task automatic setup();
      m_fip.reset();
      m_board.setSpeed(m_speed);
      m_cycle = new;
      m_cycle.setSpeed( m_speed );
      m_cycle.setVariant ( m_macrocycleVariant );
      m_cycle.configureMacrocycle( m_fip );
      m_fip.startMacrocycle();
      m_cycle.disableProducedVariable( 1 );
      m_cycle.setPeriodicVariables( m_fip );
      m_cycle.setMessages( m_fip );
      #100us;
   endtask // setup


   virtual task automatic validateVariables();
      automatic FipFrame f;
      retrieveVarsFromHMQ( 3, 1 );

      f = m_agents[1].getConsumedVar( 'h1 );

      msg(0, "Checking if MPS status indicates a non-refreshed variable");
      if ( f.getMPSStatus() != 'h00 )
	begin
	   fail($sformatf("Expected MPS status in frame produced for agent 1 == 0x0, got 0x%x instead", f.getMPSStatus()  ) );
	end
      
      
      expectConsumedVariable( 2, 'h1 );
      expectConsumedVariable( 3, 'h7b );
      expectVariablePayload( 0, 1 ); // first comes agent 1
      expectVariablePayload( 1, 2 ); // then comes agent 2
      expectVariablePayload( 2, 3 ); // ...then agent 3
      expectMessagePayload( 0, 1 );
   endtask // validateVariables

   virtual task automatic validateReport();
      readReport();
      expectTxErrors(0);
      expectRxOk(6);
      expectRxErrors(0);
      expectMPSRxErrors(0);
      expectMPSTxErrors(1);
   endtask // validateREport
   
endclass // FipTC_APP_VariableNotUpdated

class FipTC_APP_VariableUpdatedMultipleTimes extends FipTestCaseBase;

   function new;
      super.new();
   endfunction // new
   
   function automatic string getName();
      return "Application accessing multiple times the data to produce";
   endfunction // getName


   virtual task automatic run();
      automatic MQueueMessage msg_var3_payload = new ( '{ 'h8 , 'h0 , 'h2 , 'h7c , 'h3020100 , 'h7060504 , 'hb0a0908 , 'hf0e0d0c , 'h13121110 
							  , 'h17161514 , 'h1b1a1918 , 'h1f1e1d1c , 'h23222120 , 'h27262524 , 'h2b2a2928 
							  , 'h2f2e2d2c , 'h33323130 , 'h37363534 , 'h3b3a3938 , 'h3f3e3d3c , 'h43424140 
							  , 'h47464544 , 'h4b4a4948 , 'h4f4e4d4c , 'h53525150 , 'h57565554 , 'h5b5a5958 
							  , 'h5f5e5d5c , 'h63626160 , 'h67666564 , 'h6b6a6968 , 'h6f6e6d6c , 'h73727170 
							  , 'h77767574 , 'h7b7a7978 });

      automatic MockTurtleDriver mt = m_fip.MockTurtle();
      automatic MQueueHost hmq = mt.HMQ();
      const int n_attempts = 50;
      
      super.run();

      for(int i = 0; i< n_attempts; i++)
	begin
	   msg(2, $sformatf("Sending Variable 3 (0xb4) payload (%-0d/%-0d attempts)", i, n_attempts) );
	   hmq.sendMessage (`MSTRFIP_HMQ_I_CPU1_CMD, msg_var3_payload );
	end
   endtask // run
   

   virtual task automatic validateVariables();
      retrieveVarsFromHMQ( 3, 1 );
      expectConsumedVariable( 1, 'h3f );
      expectConsumedVariable( 2, 'h1 );
      expectConsumedVariable( 3, 'h7b );
      expectVariablePayload( 0, 1 ); // first comes agent 1
      expectVariablePayload( 1, 2 ); // then comes agent 2
      expectVariablePayload( 2, 3 ); // ...then agent 3
      expectMessagePayload( 0, 1 );
   endtask // validateVariables

   virtual task automatic validateReport();
      readReport();
      expectTxErrors(0);
      expectRxOk(6);
      expectRxErrors(0);
      expectMPSRxErrors(0);
      expectMPSTxErrors(0);
   endtask // validateREport
   
endclass 


// Reception of incomplete RP_DAT frame (stopped in the middle of a frame)
class FipTC_RP_DAT_Duplicated extends FipTestCaseBase;

   function new;
      super.new();
   endfunction // new
   
   function automatic string getName();
      return "Reception of two consecutive RP_DATs (before the masters sends an ID_DAT)";
   endfunction // getName

   task automatic setup();
      automatic FipAgentTransaction trans;

      super.setup();
      trans = m_agents[1].getTransaction('h02);
      trans.failureMode = AFM_DUPLICATE_RP_DAT;
   endtask // setup

   virtual task automatic validateVariables();
      retrieveVarsFromHMQ( 3, 1 );
      expectConsumedVariable( 1, 'h3f );
      expectConsumedVariable( 2, 'h1 );
      expectConsumedVariable( 3, 'h7b );
      expectVariablePayload( 0, 1 ); // first comes agent 1
      expectVariablePayload( 1, 2, `MSTRFIP_FRAME_ERR ); // then comes agent 2
      expectVariablePayload( 2, 3, `MSTRFIP_FRAME_ERR ); // ...then agent 3
      expectMessagePayload( 0, 1 );
   endtask // validateVariables

   virtual task automatic validateReport();
      readReport();
      expectTxErrors(0);
      expectRxOk(4);
      expectRxErrors(2);
      expectMPSRxErrors(0);
      expectMPSTxErrors(0);
   endtask // validateREport
   
endclass // FipTC_RP_DAT_PrematureEnd

// a variety of RP_MSG related tests
class FipTC_RP_MSG_Error extends FipTestCaseBase;

   int m_variant;
   
   
   function new(int variant);
      super.new();
      m_variant = variant;
   endfunction // new
   
   function automatic string getName();
      case(m_variant)
	0:
	  return 	"Reception of RP_MSG with unknown CTRL byte";
	1:
	  return 	"Reception of RP_MSG with CTRL byte of RP_DAT";
	2:
	  return 	"Reception of RP_MSG, not followed by RP_FIN";
	3:
	  return 	"Reception of RP_MSG, followed by RP_FIN that comes after the silence time";
	4:
	  return        "Reception of RP_FIN, with unknown CTRL byte";
	5:
	  return 	"Reception of RP_FIN with CTRL byte of RP_DAT";
	6:
	  return 	"Reception of RP_FIN with CTRL byte of RP_MSG";
      endcase // case (m_variant)
   endfunction // getName

   task automatic setup();
      automatic FipAgentTransaction trans;
      
      trans = m_agents[1].getTransaction('h02);
      
      super.setup();
      case(m_variant)
	0:
	     trans.producedMsg.forceControlField( 1, 'hbb ); // dummy CTRL byte
	1:
	     trans.producedMsg.forceControlField( 1, FT_RP_DAT ); // dummy CTRL byte
	2:
	     trans.failureMode = AFM_DROP_RP_FIN;
	3:
	  begin
	     trans.failureMode = AFM_OVERRIDE_RP_FIN;
	     trans.rpFinFrame = FipFrame::makeRP_FIN();
	     trans.rpFinFrame.forceTxTimeOffset(1, m_agents[1].getSilenceTime() );
	  end
	4:
	  begin
	     trans.failureMode = AFM_OVERRIDE_RP_FIN;
	     trans.rpFinFrame = FipFrame::makeRP_FIN();
	     trans.rpFinFrame.forceControlField(1, 'haa );
	  end
	5:
	  begin
	     trans.failureMode = AFM_OVERRIDE_RP_FIN;
	     trans.rpFinFrame = FipFrame::makeRP_FIN();
	     trans.rpFinFrame.forceControlField(1, FT_RP_DAT );
	  end
	6:
	  begin
	     trans.failureMode = AFM_OVERRIDE_RP_FIN;
	     trans.rpFinFrame = FipFrame::makeRP_FIN();
	     trans.rpFinFrame.forceControlField(1, FT_RP_MSG_NOACK );
	  end
	  
      endcase // case (m_variant)
      

	  
   endtask // setup

   virtual task automatic validateVariables();
      case(m_variant)
	0, 1:
	  begin
	  retrieveVarsFromHMQ( 3, 0 );

	     expectConsumedVariable( 1, 'h3f );
	     expectConsumedVariable( 2, 'h1 );
	     expectConsumedVariable( 3, 'h7b );
	     expectVariablePayload( 0, 1 ); // first comes agent 1
	     expectVariablePayload( 1, 2 ); // then comes agent 2
	     expectVariablePayload( 2, 3 ); // ...then agent 3
	  end // case: 0
	
	2: // no RP FIN -> timeout
	  begin
	     retrieveVarsFromHMQ( 3, 1 );
	     expectConsumedVariable( 1, 'h3f );
	     expectConsumedVariable( 2, 'h1 );
	     expectConsumedVariable( 3, 'h7b );
	     expectVariablePayload( 0, 1 ); // first comes agent 1
	     expectVariablePayload( 1, 2 ); // then comes agent 2
	     expectVariablePayload( 2, 3 ); // ...then agent 3
	     expectMessagePayload( 0, 1, `MSTRFIP_FRAME_TMO );
	  end
	
	3, 4, 5, 6:
	  begin
	     retrieveVarsFromHMQ( 3, 1 );
	     expectConsumedVariable( 1, 'h3f );
	     expectConsumedVariable( 2, 'h1 );
	     expectConsumedVariable( 3, 'h7b );
	     expectVariablePayload( 0, 1 ); // first comes agent 1
	     expectVariablePayload( 1, 2 ); // then comes agent 2
	     expectVariablePayload( 2, 3 ); // ...then agent 3
	     expectMessagePayload( 0, 1, `MSTRFIP_FRAME_BAD_CTRL );
	  end
      endcase // case (m_variant)
      
	
endtask // validateVariables
   

   virtual task automatic validateReport();
      readReport();

      case ( m_variant )
	0, 1:
	  begin
	     expectTxErrors(0);
	     expectRxOk(6);
	     expectRxErrors(0);
	     expectMPSRxErrors(0);
	     expectMPSTxErrors(0);
	  end
	2:
	  begin
	     expectTxErrors(0);
	     expectRxOk(5);
	     expectRxErrors(0);
	     expectRxTimeouts(1);
	     expectMPSRxErrors(0);
	     expectMPSTxErrors(0);
	  end
      endcase // case ( m_variant )
      
   endtask // validateREport
   
endclass // FipTC_RP_DAT_PrematureEnd



module main;

   parameter g_fast_load_cpu_firmware = 0;
   
   reg rst_n = 0;
   reg clk_sys = 0;

   always #5ns clk_sys <= ~clk_sys;
   
   initial begin
      repeat(20) @(posedge clk_sys);
      rst_n = 1;
   end

   // Board and FIP bus instantiation
   pulldown(fip_bus.p);
   pulldown(fip_bus.n);
      
   FIPBus fip_bus();
   FieldriveFPGA fieldrive();

   FipPHY U_Agent1PHY ( fip_bus.p, fip_bus.n );
   FipPHY U_Agent2PHY ( fip_bus.p, fip_bus.n );   
   FipPHY U_Agent3PHY ( fip_bus.p, fip_bus.n );
   FipPHY U_MonitorPHY ( fip_bus.p, fip_bus.n );
   FipPHY U_VirtualMasterPHY ( fip_bus.p, fip_bus.n );

   MasterfipMezzanine U_Mezzanine ( fieldrive, fip_bus );
   
   // the Device Under Test   
   spec_masterfip_mt
     #(
       .g_simulation(1),
       .g_sim_bypass_gennum(1)
       )
   
   DUT (
	.l_rst_n_i (rst_n),
  
	.sim_clk_sys_i(clk_sys),
	
	.speed_b0_i(fieldrive.speed[0]),
	.speed_b1_i(fieldrive.speed[1]),
	.ext_sync_i(fieldrive.ext_sync),

	.fd_rxcdn_i(fieldrive.fd_rxcdn),
	.fd_rxd_i  (fieldrive.fd_rxd),
	.fd_txer_i (fieldrive.fd_txer),
	.fd_wdgn_i (fieldrive.fd_wdgn),
	.fd_rstn_o (fieldrive.fd_rstn),
	.fd_txck_o (fieldrive.fd_txck),
	.fd_txd_o  (fieldrive.fd_txd),
	.fd_txena_o (fieldrive.fd_txena),

	.sim_wb_i(U_WBMaster.out),
	.sim_wb_o(U_WBMaster.in)
	);

   
    IVHDWishboneMaster U_WBMaster
     (
      .clk_i(DUT.clk_100m_sys),
      .rst_n_i(DUT.rst_n_sys)
      );

   task automatic checkSDBSignature( CBusAccessor acc );
      automatic uint64_t sig;
      automatic Logger l = Logger::get();

      
      const uint32_t SDB_ID = 'h5344422d;
      acc.read(0, sig);

      l.startTest("Check Wishbone communication");
      l.msg(0, $sformatf("SDB Signature: 0x%x", sig));

      if(sig != SDB_ID)
	l.fail($sformatf("Incorrect SDB signature (got 0x%x, expected 0x%x)", sig, SDB_ID));
      else
      l.pass();
      
   endtask // check_sdb_signature

   task automatic checkCPUCommunication(MockTurtleDriver drv);
      automatic Logger l = Logger::get();
      automatic MQueueHost hmq = drv.HMQ();
      automatic mstrfip_trtlmsg_hdr request = new (MSTRFIP_CMD_GET_BITRATE, 0);
      automatic mstrfip_hw_speed_trtlmsg response = new;
      automatic int status;

      hmq.sendAndReceiveSync(`MSTRFIP_HMQ_I_CPU0_CMD, request, `MSTRFIP_HMQ_O_RESP_CPU0_CMD, response, 100, status );

      if(status < 0)
	begin
	   l.fail("No response from CPU 0 within 100us");
	   return;
	end else begin
	   if ( response.trtl_hdr.id != MSTRFIP_REP_BITRATE )
	     begin
		l.fail($sformatf("  Incorrect response from the RT, expected 0x10, got 0x%x", response.trtl_hdr.id ) );
		return;
	     end
	end

      l.pass();
   endtask

   
    task automatic loadCPUFirmwares(MockTurtleDriver drv);
       automatic Logger l = Logger::get();
       automatic NodeCPUControl csr = drv.CPUs();
       const string cpu0_fw = "../../../masterfip/rt/ba/masterfip-rt-ba.bin";
       const string cpu1_fw = "../../../masterfip/rt/cmd/masterfip-rt-cmd.bin";

       l.startTest("Load CPU Firmwares & check CPU communication");
       
       if (!g_fast_load_cpu_firmware)
	 begin
	    l.msg(0, $sformatf("Loading '%s' to CPU0", cpu0_fw) );
	    csr.load_firmware_bin(0, cpu0_fw);
	    l.msg(0, $sformatf("Loading '%s' to CPU1", cpu1_fw) );
	    csr.load_firmware_bin(1, cpu1_fw);
	 end


       l.msg(0, "Resetting CPUs");
       csr.reset_core(0, 0);
       csr.reset_core(1, 0);

       #100us; // give some time for the firmware to boot up
       
       checkCPUCommunication(drv);
    endtask // loadCPUFirmwares
   
   task automatic createTests( ref FipTestCaseBase tests[$] );
      int speeds[] = { 2500000, 1000000, 31250 };
      
      automatic FipTCFullMacrocycle fc;

      for (int i = 0; i < 2; i++)
	begin
	   fc = new;
	   fc.setSpeed( speeds[i] );
	   fc.setMacrocycleVariant( 0 );
	   tests.push_back(fc);

	   
	   fc = new;
	   fc.setSpeed( speeds[i] );
	   fc.setMacrocycleVariant( 1 );
	   tests.push_back(fc); 
	end // for (int i = 0; i < 3; i++)
       
      begin
	 automatic FipTC_RP_DAT_Wrong_Ctrl wc1 = new( 'haa ); 
	 automatic FipTC_RP_DAT_Wrong_Ctrl wc2 = new( FT_RP_MSG_NOACK ); 
	 automatic FipTC_RP_DAT_Wrong_Ctrl wc3 = new( FT_RP_FIN ); 
	 automatic FipTC_RP_DAT_Wrong_Ctrl wc4 = new( FT_RP_ACK_EVEN);

	 automatic FipTC_RP_DAT_Wrong_PDU wp1 = new( 'hbb );  // unknown PDU type
	 automatic FipTC_RP_DAT_Wrong_PDU wp2 = new( 'h50 );  // PDU type of SMMPS
	 automatic FipTC_RP_DAT_Wrong_Preamble wp3 = new(0); // incorrect preamble
	 automatic FipTC_RP_DAT_Wrong_Preamble wp4 = new(1); // preamble truncated to 5 bits
	 automatic FipTC_RP_DAT_Wrong_Preamble wp5 = new(2); // preamble truncated to 3 bits

	 automatic FipTC_RP_DAT_TurnaroundViolation wtv = new;
	 automatic FipTC_RP_DAT_SentAfterSilence sas = new;
	 automatic FipTC_RP_DAT_WrongFES feserr1 = new(0);  
	 automatic FipTC_RP_DAT_WrongFES feserr2 = new(1);  
	 automatic FipTC_RP_DAT_WrongCRC crcerr = new;  

	 automatic FipTC_RP_DAT_IncorrectLength lenerr1 = new(0);  
	 automatic FipTC_RP_DAT_IncorrectLength lenerr2 = new(1);  
	 automatic FipTC_RP_DAT_IncorrectLength lenerr3 = new(2);  
	 automatic FipTC_RP_DAT_IncorrectLength lenerr4 = new(3);  
	 automatic FipTC_RP_DAT_IncorrectLength lenerr5 = new(4);  
	 
	 automatic FipTC_RP_DAT_WrongMPS mps1 = new(0);  
	 automatic FipTC_RP_DAT_WrongMPS mps2 = new(1);  

	 automatic FipTC_RP_DAT_PrematureEnd preend = new;  

	 automatic FipTC_RP_DAT_FieldriveError fderr1 = new(0);
	 automatic FipTC_RP_DAT_FieldriveError fderr2 = new(1);
	 automatic FipTC_APP_VariableNotUpdated apperr1 = new;
	 automatic FipTC_APP_VariableUpdatedMultipleTimes apperr2 = new;


	 automatic FipTC_RP_DAT_Duplicated duperr1 = new;
	 automatic FipTC_RP_MSG_Error msgerr1 = new(0);
	 automatic FipTC_RP_MSG_Error msgerr2 = new(1);
	 automatic FipTC_RP_MSG_Error msgerr3 = new(2);
	 automatic FipTC_RP_MSG_Error msgerr4 = new(3);
	 automatic FipTC_RP_MSG_Error msgerr5 = new(4);
	 automatic FipTC_RP_MSG_Error msgerr6 = new(5);
	 automatic FipTC_RP_MSG_Error msgerr7 = new(6);
	  
	 tests.push_back(msgerr1);
	 tests.push_back(msgerr2);
	 tests.push_back(msgerr3);
//	 tests.push_back(msgerr4);
	 tests.push_back(msgerr5);
	 tests.push_back(msgerr6);
	 tests.push_back(msgerr7);
	 
	 tests.push_back(duperr1);
	 tests.push_back(apperr1);
	 tests.push_back(apperr2);
	 tests.push_back(fderr1);
	 tests.push_back(fderr2);
	 tests.push_back(wc1);
	 tests.push_back(wc2);
	 tests.push_back(wc3);
	 tests.push_back(wc4);
	 tests.push_back(wp1);
	 tests.push_back(wp2);
	 tests.push_back(wtv);
	 tests.push_back(sas);
	 tests.push_back(feserr1);
	 tests.push_back(feserr2);
	 tests.push_back(crcerr);
	 tests.push_back(lenerr1);
	 tests.push_back(lenerr2);
	 tests.push_back(lenerr3);
	 tests.push_back(lenerr4);
	 tests.push_back(lenerr5);
	 tests.push_back(mps1);
	 tests.push_back(mps2);
	 tests.push_back(preend);

	 
      end
      
	 
			

      

   endtask // createTests
   
   initial begin
      automatic CWishboneAccessor acc = U_WBMaster.get_accessor();
      automatic MockTurtleDriver drv = new(acc, 'h20000);
      automatic MasterfipDriver fip = new (drv);
      automatic Logger l = Logger::get();
      automatic FipTestCaseBase tests[$];


      
      #10us;
      
      U_Mezzanine.setSpeed( 2500000 );

      acc.set_mode (PIPELINED);
      
      checkSDBSignature(acc);

      drv.init();
      loadCPUFirmwares(drv);

      createTests( tests );

      
      
      foreach( tests[i] )
	begin
	   automatic FipTestCaseBase test = tests[i];
	   
	   l.startTest( test.getName() );

	   test.setDriver ( fip );
	   test.attachBoard( U_Mezzanine );
	   test.configureAgents( U_MonitorPHY, U_Agent1PHY, U_Agent2PHY, U_Agent3PHY );
	   test.setup();
	   test.run();

	   while(!test.isFinished())
	     begin
		fip.handleDebugMessages();
		fip.pollPayloads();
		#2us;
	     end
	   

	   test.validate();
	   l.pass();
      end

       l.writeTestReport ( "report.txt" );

      $finish;

    end
 

endmodule // main


