/*
 * This program source code file is part of MasterFip project.
 *
 * Copyright (C) 2013-2017 CERN
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, you may find one here:
 * http://www.gnu.org/licenses/old-licenses/gpl-2.0.html
 * or you may search the http://www.gnu.org website for the version 2 license,
 * or you may write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 */

/* 
   masterfip_driver.svh - Virtual MasterFIP device driver, on top of the Mock
   Turtle library. Provides minimum API functionality necessary for verification
   of the RT code.
*/

`ifndef __MASTERFIP_DRIVER_INCLUDED
 `define __MASTERFIP_DRIVER_INCLUDED

 `include "logger.svh"
 `include "mock_turtle_driver.svh"
 `include "serializable.svh"
 `include "masterfip_common.svh"
 `include "fip_frame.svh"


// copies a number of bytes from a word-aligned buffer (reported to HMQ by the MasteFIP)
// to an array, swapping the endianess.
function automatic void copyLEBytes( ByteBuffer data, ref uint8_t dst[$], input int count );
   int i;
   
   for(i=0;i<(count+3)/4;i++)
     begin
	int remaining = (count - i * 4) >= 4 ? 4 : (count - i * 4);
	uint32_t w = data.getWord();

	for(int j = 0; j < remaining;j++)
	  begin
	     uint8_t b = (w >> (8*j)) & 'hff;
	     dst.push_back(b);
	  end
     end
endfunction // copyLEBytes

class MasterfipVariablePayload extends Serializable;
   uint32_t error;
   uint8_t payload[$];

   virtual function automatic void deserialize( ByteBuffer data );
      automatic uint32_t key = data.getWord();
      automatic uint32_t payload_bsz = data.getWord();

      error = data.getWord();
      copyLEBytes( data, payload, payload_bsz);
   endfunction // deserialize

   virtual function automatic FipFrame asFipFrame( FipFrameControl control );
      automatic FipFrame f = new;
      automatic uint8_t raw[$] = '{ control };
      for(int i = 0; i < payload.size(); i++)
	raw.push_back(payload[i]);
      f.deserializeBytes(raw);
      return f;
   endfunction // asFipFrame
   
   
endclass // MasterfipVariablePayload

class MasterfipHostTLVMessage #( type FrameType = MasterfipVariablePayload ) extends Serializable;

   bit 	   m_useMessageHack;

   function new;
      m_useMessageHack = 0;
   endfunction // deserialize
   
   
   FrameType m_payloads[$];
   
   virtual function automatic void  deserialize( ByteBuffer data );
      automatic uint32_t tlv_type, tlv_len;
      automatic uint32_t tlv_count;
      
/* -----\/----- EXCLUDED -----\/-----
      for (int i = 0; i < (data.size()+3)/4;i++)
	$display("%x: %x", i, data.getWord());
 -----/\----- EXCLUDED -----/\----- */
  
      data.reset();

      // skip header
      data.getWord();
      tlv_count = data.getWord();
      data.getWord();
      

      for(int i = 0; i < tlv_count; i++)
	begin
	   automatic int startPos, endPos;
	   startPos = data.getPos() & 'hffc;
	   
	   tlv_type = data.getWord();
	   tlv_len = data.getWord();

/* -----\/----- EXCLUDED -----\/-----
	   $display("parse tlv %d %d\n", tlv_type, tlv_len );
 -----/\----- EXCLUDED -----/\----- */
	   
	   
	   if(tlv_len == 0)
	     return;
		    
	   case(tlv_type)
	     0: begin // DATA TLV
		
		automatic FrameType p = new;
//		$display("DATA");
		p.deserialize(data);
		m_payloads.push_back(p);

/* -----\/----- EXCLUDED -----\/-----
		if(m_useMessageHack && !p.error)
		  begin
		     $display("hack in action!");
		     tlv_len += 5;
		  end
 -----/\----- EXCLUDED -----/\----- */

		
	     end // case: 0
	     1: begin // IRQ TLV : skip for the moment
		automatic int i;
//		$display("IRQ");
		
		for(i=0;i<6;i++) data.getWord();
		
	     end
	     default:
	       $error("WRONG TLV %d %d", tlv_type, tlv_len);
	     
	   endcase // case (tlv_type)

	   endPos = data.getPos() & 'hffc;

//	   $display("start_pos %d tlv_len %d", startPos, tlv_len );
	   

	   
	   
	   data.setPos( startPos + tlv_len * 4 );
	   
	   
//	   $display("d %d tlvLen %d type %d", (endPos - startPos) / 4, tlv_len, tlv_type);
	end // while (data.getPos() < data.size() )

//      $display("\n\n\n\nFrames: %d irqs: %d\n\n\n\n", frames, irqs);
      
   endfunction // deserialize

   
   
endclass // mstrfip_trtlmsg_hdr



class MasterfipDriver;

   protected MockTurtleDriver m_mt;

   protected const time c_hmq_timeout = 2000;
   
   
   int 		    m_bitrate;
   
   static const uint32_t mstrfip_bit_nstime[] = '{30769, 1000, 400};
   static const uint32_t mstrfip_tr_nstime[] = '{532000, 10000, 6000};
   static const uint32_t mstrfip_ts_nstime[] = '{4312000, 178000, 116000};
   static const int MSTRFIP_trtl_CPU_TICKS_NS = 10;
   
   function new(MockTurtleDriver mt);
      m_mt = mt;
   endfunction // repeat

   function MockTurtleDriver MockTurtle();
      return m_mt;
   endfunction // MockTurtle

   task automatic reset();
      automatic NodeCPUControl csr;
      csr = m_mt.CPUs();
      csr.reset_core( 0, 1 );
      csr.reset_core( 1, 1 );
      csr.reset_core( 0, 0 );
      csr.reset_core( 1, 0 );
      #100us;
      clearPayloads();
      
   endtask // reset
   
   
   task automatic init();
      automatic Logger l = Logger::get();
      automatic NodeCPUControl csr;
      const uint32_t MSTRFIP_APP_ID= 'hf1dc03e;
      automatic uint32_t id, core_count, mem0_size, mem1_size;

      l.startTest("Initialize MT drivers");
      
      m_mt.init();
      csr = m_mt.CPUs();
      id = csr.getAppID();

      l.msg(0, $sformatf("Read App ID: 0x%x", id));
      if ( id != MSTRFIP_APP_ID )
	begin
	   l.fail($sformatf("  Incorrect App ID: 0x%x, expected %x", id, MSTRFIP_APP_ID));
	   return;
	end

      core_count = csr.getCoreCount();

      l.msg(0, $sformatf("Read CPU core count: %d", core_count));
      if ( core_count != 2 )
	begin
	   l.fail($sformatf("Incorrect CPU cores number: %d, expected 2", core_count));
	   return;
	end

      l.pass();
   endtask // init
   

   
   task automatic getBitrate(ref int bitrate, ref int status);
      MQueueHost hmq = m_mt.HMQ();
      mstrfip_trtlmsg_hdr request = new (MSTRFIP_CMD_GET_BITRATE, 0);
      mstrfip_hw_speed_trtlmsg response = new;
      hmq.sendAndReceiveSync(`MSTRFIP_HMQ_I_CPU0_CMD, request, `MSTRFIP_HMQ_O_RESP_CPU0_CMD, response, 100, status );
      if( status )
	return;
      
      bitrate = response.bitrate;
   endtask // getBitrate

   task automatic getVersion(mstrfip_rt_version_trtlmsg response, ref int status);
      MQueueHost hmq = m_mt.HMQ();
      mstrfip_trtlmsg_hdr request = new (MSTRFIP_CMD_GET_BITRATE, 0);
      hmq.sendAndReceiveSync(`MSTRFIP_HMQ_I_CPU0_CMD, request, `MSTRFIP_HMQ_O_RESP_CPU0_CMD, response, 100, status );
      if( status )
	return;
      
   endtask // getVersion
   
   
   task automatic startMacrocycle();
      automatic MQueueHost hmq = m_mt.HMQ();
      automatic MQueueMessage msg_ba_start = new ('{'h7, 'h0, 'h1} );
      automatic Logger l = Logger::get();
      
      l.msg(2,"Sending BA_START");
      hmq.sendMessage (`MSTRFIP_HMQ_I_CPU1_CMD, msg_ba_start );
      
      endtask // startMacrocycle

   task automatic getReport( mstrfip_report report );
      automatic Logger l = Logger::get();
      automatic MQueueHost hmq = m_mt.HMQ();
      automatic MQueueMessage msg_get_report = new ('{'hc, 'h0, 'h0} );
      automatic int status;

      
      hmq.sendAndReceiveSync (`MSTRFIP_HMQ_I_CPU1_CMD, msg_get_report, `MSTRFIP_HMQ_O_RESP_CPU1_CMD, report, 2000, status );
      if( status )
	begin
	   l.fail("No response for GET_REPORT message");
	   return;
	end
   endtask // getReport
   
   task automatic readPayloads( int slot, ref MasterfipVariablePayload payloads[$], input int count, input int useMessageHack, ref int status, input real timeout = 10000 );
      automatic MQueueHost hmq = m_mt.HMQ();


      while(count > 0)
	begin
	   automatic MasterfipHostTLVMessage tlv_msg = new;
	   tlv_msg.m_useMessageHack = useMessageHack;
	   

	   
	   hmq.receiveMessage(  slot, tlv_msg, status, timeout );

	   if( status < 0 )
	     return;

	   for (int i=0;i<tlv_msg.m_payloads.size();i++)
	     begin
		
		payloads.push_back(tlv_msg.m_payloads[i]);
		count--;
		
	     end
	end // while (count > 0)
      
      
   endtask // getVariablePayload
      
   task automatic getPresentList( ref mstrfip_presence_report report, ref int status, input real timeout = 1000 );
      automatic MQueueHost hmq = m_mt.HMQ();
      automatic mstrfip_trtlmsg_hdr request = new (MSTRFIP_CMD_GET_PRESENT_LIST, 0);

      hmq.sendAndReceiveSync(`MSTRFIP_HMQ_I_CPU1_CMD, request, `MSTRFIP_HMQ_O_RESP_CPU1_CMD, report, 100, status );


   endtask // getMessagePayloads


   typedef MasterfipVariablePayload MasterfipVariablePayloads[$];
   MasterfipVariablePayloads m_msgPayloads, m_varPayloads;
 
   
   task automatic pollPayloads();
      int status;
      readPayloads( `MSTRFIP_HMQ_O_APP_PER_VAR, m_varPayloads, 1, 0, status, 1 );
      readPayloads( `MSTRFIP_HMQ_O_APP_APER_MSG, m_msgPayloads, 1, 1, status, 1 );
   endtask // pollPayloads

   task automatic clearPayloads();
      m_varPayloads= '{};
      m_msgPayloads= '{};
   endtask // clearPayloads

   function automatic MasterfipVariablePayloads getMessagePayloads();
      return m_msgPayloads;
   endfunction // getMessagePayloads

   function automatic MasterfipVariablePayloads getVariablePayloads();
      return m_varPayloads;
   endfunction // getVariablePayloads

   task automatic handleDebugMessages();
      automatic NodeCPUControl csr;
      csr = m_mt.CPUs();
      csr.update();
   endtask // handleDebugMessages
   
endclass // MasterfipDriver


`endif
